unit UFormMain;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ExtCtrls,
  ComCtrls, UCpu, UMachine, UInstructionWriter, syncobjs;

type

  { TFormMain }

  TFormMain = class(TForm)
    ButtonStart: TButton;
    ButtonStop: TButton;
    Image1: TImage;
    Label1: TLabel;
    Label2: TLabel;
    ListViewMemory: TListView;
    ListViewRegisters: TListView;
    Memo1: TMemo;
    Timer1: TTimer;
    TimerDraw: TTimer;
    procedure ButtonStartClick(Sender: TObject);
    procedure ButtonStopClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure ListViewMemoryData(Sender: TObject; Item: TListItem);
    procedure ListViewRegistersData(Sender: TObject; Item: TListItem);
    procedure Memo1KeyPress(Sender: TObject; var Key: char);
    procedure ScreenRepaing;
    procedure PaintBox1Resize(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure TimerDrawTimer(Sender: TObject);
  private
    procedure ReloadMemoryDump;
    procedure ReloadRegisterDump;
    procedure LoadProgram;
  public
    Machine: TMachine;
    InstructionWriter: TInstructionWriter;
  end;

var
  FormMain: TFormMain;

implementation

{$R *.lfm}

const
  ItemsPerLine = 16;

{ TFormMain }

procedure TFormMain.ButtonStartClick(Sender: TObject);
begin
  Machine.Cpu.Start;
end;

procedure TFormMain.ButtonStopClick(Sender: TObject);
begin
  Machine.Cpu.Stop;
end;

procedure TFormMain.FormCreate(Sender: TObject);
begin
  Machine := TMachine.Create;
  InstructionWriter := TInstructionWriter.Create;
  InstructionWriter.Cpu := Machine.Cpu;
end;

procedure TFormMain.FormDestroy(Sender: TObject);
begin
  InstructionWriter.Free;
  Machine.Free;
end;

procedure TFormMain.FormShow(Sender: TObject);
begin
  LoadProgram;
  PaintBox1Resize(Self);
end;

procedure TFormMain.ListViewMemoryData(Sender: TObject; Item: TListItem);
var
  Line: string;
  I: Integer;
begin
  if Item.Index < Machine.MemorySize div ItemsPerLine then begin
    Line := '';
    for I := 0 to ItemsPerLine - 1 do
      Line := Line + IntToHex(PByte(Machine.Memory + Item.Index * ItemsPerLine + I)^, 2) + ' ';
    Item.Caption := IntToHex(Item.Index * ItemsPerLine, 8);
    Item.SubItems.Add(Line);
  end;
end;

procedure TFormMain.ListViewRegistersData(Sender: TObject; Item: TListItem);
begin
  if Item.Index < Length(Machine.Cpu.Registers) + 2 then begin
    if Item.Index = 0 then begin
      Item.Caption := 'IP';
      Item.SubItems.Add(IntToHex(Machine.Cpu.IP, 16));
    end else
    if Item.Index = 1 then begin
      Item.Caption := 'SP';
      Item.SubItems.Add(IntToHex(Machine.Cpu.SP, 16));
    end else begin
      Item.Caption := 'R' + IntToStr(Item.Index - 2);
      Item.SubItems.Add(IntToHex(Machine.Cpu.Registers[Item.Index - 2].D, 16));
    end;
  end;
end;

procedure TFormMain.Memo1KeyPress(Sender: TObject; var Key: char);
begin
  Machine.LockInput.Acquire;
  Machine.InputBuffer := Machine.InputBuffer + Key;
  Machine.LockInput.Release;
end;

procedure TFormMain.ScreenRepaing;
var
  B: Byte;
  P: Integer;
begin
  with Machine, Screen do begin
    if ChangedAreaFrom < ChangedAreaTo then begin
      if (Image1.Picture.Bitmap.Width <> Image1.Width) or
      (Image1.Picture.Bitmap.Height <> Image1.Height) then
        Image1.Picture.Bitmap.SetSize(Image1.Width, Image1.Height);
      Image1.Picture.Bitmap.BeginUpdate(True);
      P := ChangedAreaFrom;
      while P <= ChangedAreaTo do begin
        B := PByte(Memory + MemoryBase + P)^;
        Image1.Canvas.Pixels[P mod Size.X, P div Size.X] := B * $010101;
        Inc(P);
      end;
      ChangedAreaTo := MemoryBase;
      ChangedAreaFrom := MemoryBase + MemorySize;
      Image1.Picture.Bitmap.EndUpdate;
    end;
  end;
end;

procedure TFormMain.PaintBox1Resize(Sender: TObject);
begin
  Machine.Screen.Size := Point(Image1.Width, Image1.Height);
end;

procedure TFormMain.Timer1Timer(Sender: TObject);
begin
  Label1.Caption := 'Ticks: ' + IntToStr(Machine.Cpu.Ticks);
  ReloadMemoryDump;
  ReloadRegisterDump;
  Machine.LockOutput.Acquire;
  Memo1.Lines.Text := Memo1.Lines.Text + Machine.OutputBuffer;
  SetLength(Machine.OutputBuffer, 0);
  Machine.LockOutput.Release;
end;

procedure TFormMain.TimerDrawTimer(Sender: TObject);
begin
  ScreenRepaing;
end;

procedure TFormMain.ReloadMemoryDump;
begin
  ListViewMemory.Items.Count := Machine.MemorySize div ItemsPerLine;
  ListViewMemory.Refresh;
end;

procedure TFormMain.ReloadRegisterDump;
begin
  ListViewRegisters.Items.Count := Length(Machine.Cpu.Registers);
  ListViewRegisters.Refresh;
end;

procedure TFormMain.LoadProgram;
var
  R1: Byte;
  R2: Byte;
  R3: Byte;
  R4: Byte;
  LabelStart: Integer;
  LabelText: Integer;
  LabelPrint: Integer;
  LabelClearScreen: Integer;
begin
  R1 := 1;
  R2 := 2;
  R3 := 3;
  R4 := 4;
  with InstructionWriter do begin
    Init;
    // Print Hello world text
    LabelText := 200;
    Loadi(R1, LabelText);
    Loadi(R2, 12);
  LabelPrint := IP;
    DataPrefix8; LoadMem(R3, R1);
    DataPrefix8; Output(0, R3);
    Increment(R1);
    Decrement(R2);
    Test(R2);
    JumpNotZero(LabelPrint);

    // Draw to video memory
    Loadi(R1, Machine.Screen.MemoryBase);
    Loadi(R2, $10000);
    Loadi(R3, $7f);
  LabelClearScreen := IP;
    DataPrefix8; StoreMem(R1, R3);
    Increment(R1);
    Decrement(R2);
    Test(R2);
    AddrPrefix8; JumpRelNotZero(LabelClearScreen);

    // Update screen area
    Push(R1);
    Loadi(R1, 0);
    Output(4, R1);
    Loadi(R1, $10000);
    Output(5, R1);
    Pop(R1);

    // Read keyboard and print to console
    Loadi(R1, 100);
  LabelStart := IP;
    Increment(R1);
    Loadi(R2, $100);
    DataPrefix8; StoreMem(R2, R1);
    DataPrefix8; Input(R2, 0);
    DataPrefix8; Output(0, R2);
    Jump(LabelStart);
    Halt;
  IP := LabelText;
    WriteString('Hello World!');
  end;
end;

end.

